
import json
import logging
from dataclasses import dataclass
from datetime import datetime
from abc import ABC, abstractmethod
from random import shuffle, sample

import aiohttp
import validators
from discord import Embed, Colour
from bs4 import BeautifulSoup as bs4
from feedparser import FeedParserDict, parse
from markdownify import markdownify
from sqlalchemy import select, insert, delete, and_
from sqlalchemy.exc import NoResultFound
from textwrap import shorten

from mutators import registry as mutator_registry
from errors import IllegalFeed
from db import DatabaseManager, RssSourceModel, FeedChannelModel
from utils import get_rss_data, get_unparsed_feed
from api import API

log = logging.getLogger(__name__)
dumps = lambda _dict: json.dumps(_dict, indent=8)


from xml.etree.ElementTree import Element, SubElement, tostring
from feedparser import parse

class RSSItem:
    def __init__(self, title, link, description, pub_date, guid):
        self.title = title
        self.link = link
        self.description = description
        self.pub_date = pub_date
        self.guid = guid

    def __str__(self):
        return self.title

    def create_mutated_copy(self, mutators):
        pass

    def create_embed(self, sub, feed):

        title = shorten(markdownify(self.title, strip=["img", "a"]), 256)
        desc = shorten(markdownify(self.description, strip=["img"]), 4096)

        author = ""
        author_url = ""

        icon_url = ""
        thumb_url = ""

        # Replace HTML with Markdown, and shorten text.
        # author = shorten(self.source.name, 256)

        # validate urls
        # author_url = self.source.url if validators.url(self.source.url) else None
        # icon_url = self.source.icon_url if validators.url(self.source.icon_url) else None
        # thumb_url = await self.get_thumbnail_url(session)  # validation done inside func

        # Combined length validation
        # Can't exceed combined 6000 characters, [400 Bad Request] if failed.
        combined_length = len(title) + len(desc) + (len(author) * 2)
        cutoff = combined_length - 6000
        desc = shorten(desc, cutoff) if cutoff > 0 else desc

        embed = Embed(
            title=title,
            description=desc,
            timestamp=self.published,
            url=self.link if validators.url(self.link) else None,
            colour=colour
        )

        # embed.set_thumbnail(url=icon_url)
        # embed.set_image(url=thumb_url)
        # embed.set_author(url=author_url, name=author)
        # embed.set_footer(text=self.author)

        return embed


class RSSFeed:
    def __init__(self, title, link, description, language='en-gb', pub_date=None, last_build_date=None):
        self.title = title
        self.link = link
        self.description = description
        self.language = language
        self.pub_date = pub_date
        self.last_build_date = last_build_date
        self.items = []

    def add_item(self, item: RSSItem):
        if not isinstance(item, RSSItem):
            raise TypeError("item must be an instance of RSSItem")

        self.items.append(item)

    def __str__(self):
        return self.guid

    @classmethod
    def from_parsed_feed(cls, parsed_feed):
        title = parsed_feed.feed.get('title', 'No title')
        link = parsed_feed.feed.get('link', 'No link')
        description = parsed_feed.feed.get('description', 'No description')
        language = parsed_feed.feed.get('language', 'en-gb')
        pub_date = parsed_feed.feed.get('published', None)
        last_build_date = parsed_feed.feed.get('updated', None)

        feed = cls(title, link, description, language, pub_date, last_build_date)

        for entry in parsed_feed.entries:
            item_title = entry.get('title', 'No title')
            item_link = entry.get('link', 'No link')
            item_description = entry.get('description', 'No description')
            item_pub_data = entry.get('published_parsed', None)
            item_guid = entry.get('id', None) or entry.get("guid", None)

            item_published = datetime(*entry.published_parsed[0:-2]) if published_parsed else None

            item = RSSItem(item_title, item_link, item_description, item_published, item_guid)
            feed.add_item(item)

        feed.items.reverse()

        return feed



@dataclass
class RSSArticle:
    """Represents a news article, or entry from an RSS feed."""

    guid: str
    title: str | None
    description: str | None
    url: str | None
    published: datetime | None
    author: str | None
    source: object

    @classmethod
    def from_entry(cls, source, entry:FeedParserDict):
        """Create an Article from an RSS feed entry.

        Parameters
        ----------
        entry : FeedParserDict
            An entry pulled from a complete FeedParserDict object.

        Returns
        -------
        Article
            The Article created from the feed entry.
        """

        # log.debug("Creating Article from entry: %s", dumps(entry))

        published_parsed = entry.get("published_parsed")
        published = datetime(*entry.published_parsed[0:-2]) if published_parsed else None

        return cls(
            guid=entry["guid"],
            title=entry.get("title"),
            description=entry.get("description"),
            url=entry.get("link"),
            published=published,
            author = entry.get("author"),
            source=source
        )

    def mutate(self, attr: str, mutator: dict[str, str]):
        """
        Apply a mutation to a certain text attribute of
        this Article instance.
        """

        # WARN:
        # This could be really bad if the end user is able to effect the 'attr' value.
        # Shouldn't happen though.

        log.debug("applying mutator '%s'", mutator["name"])
        val = mutator["value"]

        try:
            mutator = mutator_registry.get_mutator(val)
        except ValueError as err:
            log.error(err)

        setattr(self, attr, mutator.mutate(getattr(self, attr)))
        log.debug("mutated %s, to: %s", attr, getattr(self, attr))


    async def get_thumbnail_url(self, session: aiohttp.ClientSession) -> str | None:
        """Returns the thumbnail URL for an article.

        Parameters
        ----------
        session : aiohttp.ClientSession
            A client session used to get the thumbnail.

        Returns
        -------
        str or None
            The thumbnail URL, or None if not found.
        """

        # log.debug("Fetching thumbnail for article: %s", self)

        try:
            async with session.get(self.url, timeout=15) as response:
                html = await response.text()
        except aiohttp.InvalidURL as error:
            log.error("invalid thumbnail url: %s", error)
            return None

        soup = bs4(html, "html.parser")
        image_element = soup.select_one("meta[property='og:image']")
        if not image_element:
            return None

        image_content = image_element.get("content")
        return image_content if validators.url(image_content) else None

    async def to_embed(self, session: aiohttp.ClientSession, colour: Colour) -> Embed:
        """Creates and returns a Discord Embed object from the article.

        Parameters
        ----------
        session : aiohttp.ClientSession
            A client session used to get additional article data.

        Returns
        -------
        Embed
            A Discord Embed object representing the article.
        """

        # log.debug(f"Creating embed from article: {self}")

        # Replace HTML with Markdown, and shorten text.
        title = shorten(markdownify(self.title, strip=["img", "a"]), 256)
        desc = shorten(markdownify(self.description, strip=["img"]), 4096)
        author = shorten(self.source.name, 256)

        # validate urls
        embed_url = self.url if validators.url(self.url) else None
        author_url = self.source.url if validators.url(self.source.url) else None
        icon_url = self.source.icon_url if validators.url(self.source.icon_url) else None
        thumb_url = await self.get_thumbnail_url(session)  # validation done inside func

        # Combined length validation
        # Can't exceed combined 6000 characters, [400 Bad Request] if failed.
        combined_length = len(title) + len(desc) + (len(author) * 2)
        cutoff = combined_length - 6000
        desc = shorten(desc, cutoff) if cutoff > 0 else desc

        embed = Embed(
            title=title,
            description=desc,
            timestamp=self.published,
            url=embed_url,
            colour=colour
        )

        embed.set_thumbnail(url=icon_url)
        embed.set_image(url=thumb_url)
        embed.set_author(url=author_url, name=author)
        embed.set_footer(text=self.author)

        return embed


@dataclass
class RSSFeedSource:
    """Represents an RSS Feed."""

    name: str | None
    description: str | None
    url: str | None
    icon_url: str | None
    feed: FeedParserDict

    @classmethod
    def from_parsed(cls, feed:FeedParserDict):
        """Returns a Source object from a parsed feed.

        Parameters
        ----------
        feed : FeedParserDict
            The feed used to create the Source.

        Returns
        -------
        Source
            The Source object
        """

        # log.debug("Creating Source from feed: %s", dumps(feed))

        channel = feed.get("channel", {})

        return cls(
            name=channel.get("title"),
            description=channel.get("description"),
            url=channel.get("link"),
            icon_url=feed.get("feed", {}).get("image", {}).get("href"),
            feed=feed
        )

    @classmethod
    async def from_url(cls, url: str):
        unparsed_content = await get_unparsed_feed(url)
        source = cls.from_parsed(parse(unparsed_content))
        source.url = url
        return source

    def get_latest_articles(self, max: int = 999) -> list[Article]:
        """Returns a list of Article objects.

        Parameters
        ----------
        max : int
            The maximum number of articles to return.

        Returns
        -------
        list of Article
            A list of Article objects.
        """

        # log.debug("Fetching latest articles from %s, max=%s", self, max)

        return [
            Article.from_entry(self, entry)
            for i, entry in enumerate(self.feed.entries)
            if i < max
        ]


@dataclass
class RSSFeedSource_:

    uuid: str
    name: str
    url: str
    image: str
    discord_server_id: id
    created_at: str

    @classmethod
    def from_list(cls, data: list) -> list:
        result = []

        for item in data:
            key = "discord_server_id"
            item[key] = int(item.get(key))
            result.append(cls(**item))

        return result

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


@dataclass
class DjangoDataModel(ABC):

    @staticmethod
    @abstractmethod
    def parser(item: dict) -> dict:
        """Overwrite this method to parse types."""
        return item

    @classmethod
    def from_list(cls, data: list[dict]) -> list:
        return [cls(**cls.parser(item)) for item in data]

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**cls.parser(data))


@dataclass(slots=True)
class Subscription(DjangoDataModel):

    id: int
    name: str
    url: str
    guild_id: int
    creation_datetime: datetime
    extra_notes: str
    filters: list[int]
    mutators: dict[str, list[dict]]
    embed_colour: str
    active: bool
    channels_count: int

    @staticmethod
    def parser(item: dict) -> dict:
        item["guild_id"] = int(item["guild_id"])
        item["creation_datetime"] = datetime.strptime(item["creation_datetime"], "%Y-%m-%dT%H:%M:%S.%f%z")
        item["mutators"] = {
            "title": item.pop("article_title_mutators"),
            "desc": item.pop("article_desc_mutators")
        }

        return item

    async def get_channels(self, api):
        channel_data, _ = await api.get_subscription_channels(subscription=self.id)
        return SubChannel.from_list(channel_data)


@dataclass(slots=True)
class SubChannel(DjangoDataModel):

    id: int
    channel_id: int
    subscription: int

    @staticmethod
    def parser(item: dict) -> dict:
        item["channel_id"] = int(item["channel_id"])
        item["subscription"] = int(item["subscription"])

        return item

    @property
    def mention(self) -> str:
        return f"<#{self.channel_id}>"


@dataclass(slots=True)
class SubscriptionChannel(DjangoDataModel):

    uuid: str
    id: int
    subscription: Subscription
    creation_datetime: datetime

    @staticmethod
    def parser(item: dict) -> dict:
        
        item["id"] = int(item["id"])
        item["subscription"] = Subscription.from_dict(item.pop("subscription"))
        item["creation_datetime"] = datetime.strptime(item["creation_datetime"], DATETIME_FORMAT)
        return item

    @property
    def mention(self) -> str:
        """
        Returns the `id` as a string in the discord mention format.
        """

        return f"<#{self.id}>"

    @property
    def hyperlinks_string(self) -> str:
        """"""

        api_hyperlink = f"[API]({API.CHANNEL_ENDPOINT}{self.uuid}/)"
        rss_hyperlink = f"[RSS]({self.subscription.rss_url})"
        value = f"{rss_hyperlink} · {api_hyperlink}"

        return value


@dataclass(slots=True)
class TrackedContent(DjangoDataModel):

    uuid: str
    subscription: str
    content_url: str
    creation_datetime: datetime

    @staticmethod
    def parser(item: dict) -> dict:

        item["creation_datetime"] = datetime.strptime(item["creation_datetime"], DATETIME_FORMAT)
        return item


class Functions:

    def __init__(self, bot, api_token: str):
        self.bot = bot
        self.api_token = api_token

    async def validate_feed(self, nickname: str, url: str) -> FeedParserDict:
        """Validates a feed based on the given nickname and url.

        Parameters
        ----------
        nickname : str
            Human readable nickname used to refer to the feed.
        url : str
            URL to fetch content from the feed.

        Returns
        -------
        FeedParserDict
            A Parsed Dictionary of the feed.

        Raises
        ------
        IllegalFeed
            If the feed is invalid.
        """

        # Ensure the URL is valid
        if not validators.url(url):
            raise IllegalFeed(f"The URL you have entered is malformed or invalid:\n`{url=}`")

        # Check the nickname is not a URL
        if validators.url(nickname):
            raise IllegalFeed(
                "It looks like the nickname you have entered is a URL.\n" \
                "For security reasons, this is not allowed.",
                nickname=nickname
            )

        feed_data, status_code = await get_rss_data(url)

        if status_code != 200:
            raise IllegalFeed(
                "The URL provided returned an invalid status code:",
                url=url, status_code=status_code
            )

        # Check the contents is actually an RSS feed.
        feed = parse(feed_data)
        if not feed.version:
            raise IllegalFeed(
                "The provided URL does not seem to be a valid RSS feed.",
                url=url
            )

        return feed

    async def create_new_rssfeed(self, name: str, url: str, guild_id: int) -> RSSFeed:


        log.info("Creating new Feed: %s", name)

        parsed_feed = await self.validate_feed(name, url)
        source = Source.from_parsed(parsed_feed)

        async with aiohttp.ClientSession() as session:
            data = await API(self.api_token, session).create_new_rssfeed(
                name, url, source.icon_url, guild_id
            )

        return RSSFeed.from_dict(data)

    async def delete_rssfeed(self, uuid: str) -> RSSFeed:

        log.info("Deleting Feed '%s'", uuid)

        async with aiohttp.ClientSession() as session:
            api = API(self.api_token, session)
            data = await api.get_rssfeed(uuid)
            await api.delete_rssfeed(uuid)

        return RSSFeed.from_dict(data)

    async def get_rssfeeds(self, guild_id: int, page: int, pagesize: int) -> list[RSSFeed]:
        """Get a list of RSS Feeds.

        Args:
            guild_id (int): The guild_id to filter by.

        Returns:
            list[RSSFeed]: Resulting list of RSS Feeds
        """

        async with aiohttp.ClientSession() as session:
            data, count = await API(self.api_token, session).get_rssfeed_list(
                discord_server_id=guild_id,
                page=page,
                page_size=pagesize
            )

        return RSSFeed.from_list(data), count

    async def assign_feed(
        self, url: str, channel_name: str, channel_id: int, guild_id: int
    ) -> tuple[int, Source]:
        """"""

        async with DatabaseManager() as database:
            select_query = select(RssSourceModel).where(and_(
                RssSourceModel.rss_url == url,
                RssSourceModel.discord_server_id == guild_id
            ))
            select_result = await database.session.execute(select_query)

            rss_source = select_result.scalars().one()

            insert_query = insert(FeedChannelModel).values(
                discord_server_id = guild_id,
                discord_channel_id = channel_id,
                rss_source_id=rss_source.id,
                search_name=f"{rss_source.nick} #{channel_name}"
            )

            insert_result = await database.session.execute(insert_query)
            return insert_result.inserted_primary_key.id, await Source.from_url(url)

    async def unassign_feed( self, assigned_feed_id: int, guild_id: int):
        """"""

        async with DatabaseManager() as database:
            query = delete(FeedChannelModel).where(and_(
                FeedChannelModel.id == assigned_feed_id,
                FeedChannelModel.discord_server_id == guild_id
            ))

            result = await database.session.execute(query)
            if not result.rowcount:
                raise NoResultFound
